from ..config import config
import logging
from ..image import Image2DSpec
from ..detector.parallel import Parallel2D
from .algorithm import Algorithm
import numpy as np
from contextlib import contextmanager

import enum

# from dxpy.configs import configurable

# def _inputs_verification2d(detector, phantom_spec, sinogram):
#   from ..detector.base import Detector2D
#   from ..exceptions import InputVerifacationError
#   if not isinstance(detector, Detector2D):
#     raise InputVerifacationError("Input detector is not Detector2D, {}.")
#   detector.assert_fit(sinogram)


@contextmanager
def reconstructor2d(detector: Parallel2D, image_spec: Image2DSpec,
                    method: Algorithm):
  if method.backend == 'astra':
    try:
      from ..backend.astra.reconstruction2d import ReconstructorParallel2D
      r = ReconstructorParallel2D(detector, image_spec, method)
      yield r
    except Exception as e:
      raise e
    finally:
      r.clear()
  else:
    raise ValueError("Unknown backend {}.".format(method.backend))


def reconstruction2d_single(sinogram: np.ndarray, detector: Parallel2D,
                            image: Image2DSpec, method: Algorithm):
  with reconstructor2d(detector, image, method) as r:
    return r.reconstruct(sinogram)


def reconstruction2d_batch(sinograms: np.ndarray,
                           detector: Parallel2D,
                           image_spec: Image2DSpec,
                           method: Algorithm,
                           *,
                           tqdm=None):
  nb_images = sinograms.shape[0]
  with reconstructor2d(detector, image_spec, method) as r:
    if tqdm is None:

      def tqdm(x, ascii=None, leave=None):
        return x

    results = [
        r.reconstruct(sinograms[i, ...])
        for i in tqdm(range(nb_images), ascii=True, leave=False)
    ]
  return np.array(results)


# @configurable(config.get('reconstruction'))
def reconstruction2d(sinograms,
                     detector: Parallel2D,
                     image_spec: Image2DSpec,
                     method,
                     *,
                     tqdm=None):
  """
    Args:
      sinogram: one of the following objects:
        1. 2-dimensional ndarray
        2. Sinogram2D object
        3. NDArrayOnFS object

        detector: Detector specifics,
        phantom_spec: phantom specifics
        sinogram: 2-dimensional ndarray.
    """
  if sinograms.ndim in (3, 4):
    logging.debug('Sinogram dimension is {}, use batch mode.'.format(
        sinograms.ndim))
    if sinograms.ndim == 4:
      logging.debug(
          'Last dimension of sinograms will be ignored by sinograms[:, :, :, 0].'.
          format(sinograms.ndim))
      sinograms = sinograms[:, :, :, 0]
    return reconstruction2d_batch(
        sinograms, detector, image_spec, method, tqdm=tqdm)
  else:
    logging.debug('Sinogram dimension is 3, use batch mode.')
    return reconstruction2d_single(sinograms, detector, image_spec, method)


#   _inputs_verification2d(detector, phantom_spec, sinogram)
#   vol_geom = ac.create_vol_geom(phantom_spec.shape)
#   proj_geom = ac.create_proj_geom('parallel', detector.sensor_width,
#                                   detector.nb_sensors, detector.views)
#   proj_id = ac.create_projector('linear', proj_geom, vol_geom)
#   rid, result = ac.create_reconstruction(method, proj_id, sinogram, iterations)
#   astra.data2d.clear()
#   astra.projector.clear()
#   astra.algorithm.clear()
#   return result



class Algorithm:
  def __init__(self, device, backend):
    self._device = device
    self._backend = backend

  @property
  def device(self):
    return self._device

  @property
  def backend(self):
    return self._backend


class FBP(Algorithm):
  @classmethod
  def supported_filters(cls, backend):
    if backend == 'astra':
      return [
          'ram-lak',
          'shepp-logan',
          'cosine',
          'hamming',
          'hann',
          'none',
      ]

  def __init__(self,
               device='gpu',
               backend='astra',
               filter_='ram-lak',
               filter_d=1.0):
    super().__init__(device, backend)
    self._filter = filter_
    self.filter_d = filter_d

  @property
  def filter(self):
    return self._filter



import numpy as np


class ImageSpec:
  def __init__(self, shape, dtype=np.float32):
    self._shape = shape
    self._dtype = dtype

  @property
  def shape(self):
    return self._shape

  @property
  def dtype(self):
    return self._dtype


class Image2DSpec(ImageSpec):
  ndim = 2


class Image3DSpec(ImageSpec):
  ndim = 3



from .base import Detector2D, Model
from dxl.data.core.asserts import assert_dimension
import numpy as np


class ParallelLinearModel(Model):
  name = 'linear'


class Parallel2D(Detector2D):
  @classmethod
  def create_sensors(cls, nb_sensors, sensor_width=1.0):
    """
    Helper function to create sensors, locating at center of each periods.
    """
    return np.arange(nb_sensors) * sensor_width - \
      ((nb_sensors / 2 - 0.5) * sensor_width)

  @classmethod
  def create_views(cls, nb_views, view_range=None):
    if view_range is None:
      view_range = [0.0, np.pi]
    return np.linspace(view_range[0], view_range[1], nb_views, endpoint=False)

  def __init__(self, sensors=None, views=None, model=None):
    """
    """
    self._sensors = np.array(sensors)
    self._views = np.array(views)
    assert_dimension(self._sensors, 1, 'Detector sensors')
    assert_dimension(self._views, 1, 'Detector views')
    if model is None:
      model = ParallelLinearModel()
    self._model = model

  @property
  def nb_sensors(self):
    return len(self._sensors)

  def sensor_width(self):
    return np.mean(self._sensors[1:] - self._sensors[:-1])

  def sensors(self):
    return np.array(self._sensors)

  @property
  def nb_views(self):
    return len(self._views)

  def views(self):
    return np.array(self._views)

  def model(self):
    return self._model

  def check_data_capability(self, data):
    if isinstance(data, np.ndarray):
      if data.ndim != 2 or self.nb_views != data.shape[0] or self.nb_sensors != data.shape[1]:
        msg = "Shape of sinogram {} is not consisted with detector: nb_sensors: {}, nb_views: {}."
        raise ValueError(
            msg.format(data.shape, self.nb_sensors, self.nb_views))
    else:
      raise TypeError("Unknown data type: {}.".format(type(data)))



import astra
import astra.creators as ac
import numpy as np
from dxpy.tensor.checks import assert_ndim
from dxpy.tensor.transform import maybe_unbatch
from ..config import config
from dxpy.configs import configurable


class Sinogram:
  pass


@configurable(config.get('projection'))
def projection2d(image, detector, *, method='cuda', projection_model='linear'):
  assert_ndim(image, 2, 'image')
  vol_geom = ac.create_vol_geom(image.shape)
  proj_geom = ac.create_proj_geom('parallel', detector.sensor_width,
                                  detector.nb_sensors, detector.views)
  proj_id = ac.create_projector(projection_model, proj_geom, vol_geom)
  sid, sinogram = ac.create_sino(image, proj_id, gpuIndex=0)
  astra.data2d.clear()
  astra.projector.clear()
  astra.algorithm.clear()
  return sinogram


class Projector2DParallel:
  @configurable(config.get('projection'))
  def __init__(self,
               detector,
               phantom_spec,
               *,
               method='cuda',
               projection_model='linear'):
    vol_geom = ac.create_vol_geom(phantom_spec.shape)
    proj_geom = ac.create_proj_geom('parallel', detector.sensor_width,
                                    detector.nb_sensors, detector.views)
    self.proj_id = ac.create_projector(projection_model, proj_geom, vol_geom)
    self.phan_id = astra.data2d.create('-vol', vol_geom)

  def project(self, image):
    assert_ndim(image, 2, 'image')
    astra.data2d.store(self.phan_id, image)
    sid, result = ac.create_sino(self.phan_id, self.proj_id)
    astra.data2d.delete(sid)
    return result

  def clear(self):
    astra.data2d.delete(self.phan_id)
    astra.projector.delete(self.proj_id)


def padding_pi2full(sinogram):
  """
    Padding sinogram of [0, np.pi) to full [0, 2*np.pi)
    """
  return np.concatenate([sinogram, np.flip(sinogram, 0)], axis=1)



import numpy as np
from .spec import Image2DSpec


class Image:
  def __init__(self, shape, dtype=None):
    if dtype is None:
      dtype = np.float32
    self._shape = shape
    self._dtype = dtype

  @property
  def shape(self):
    return self._shape

  @property
  def dtype(self):
    return self._dtype


class Image2D(Image):
  def __init__(self, shape, dtype=None):
    if isinstance(shape, Image2DSpec):
      shape = shape.shape
      dtype = shape.dtype
    else:
      if len(shape) != 2:
        raise ValueError("Invalid shape {} for Image2D.".format(shape))
    super().__init__(shape, dtype)


class Detector:
  def check_data_capability(self, data):
    """
    Check if data is output of this detecotr, mainly checking shape, content, etc.
    Returns: 
      None
    Raises:
      ValueError, TypeError
    """
    pass


class Detector3D(Detector):
  ndim = 3


class Detector2D(Detector):
  ndim = 2


class Model:
  name = None



default_config = {
    'backends': ['astra-python', 'astra-matlab'],
    'astra-matlab': {
        'root': None
    },
    'projection': dict(),
    'reconstruction': dict()
}
config = default_config
